import os

import streamlit.components.v1 as components

# Create a _RELEASE constant. We'll set this to False while we're developing
# the component, and True when we're ready to package and distribute it.
# (This is, of course, optional - there are innumerable ways to manage your
# release process.)
_RELEASE = True

# Declare a Streamlit component. `declare_component` returns a function
# that is used to create instances of the component. We're naming this
# function "_component_func", with an underscore prefix, because we don't want
# to expose it directly to users. Instead, we will create a custom wrapper
# function, below, that will serve as our component's public API.

# It's worth noting that this call to `declare_component` is the
# *only thing* you need to do to create the binding between Streamlit and
# your component frontend. Everything else we do in this file is simply a
# best practice.

if not _RELEASE:
    _record_viewer = components.declare_component(
        # We give the component a simple, descriptive name
        "record_viewer",
        # Pass `url` here to tell Streamlit that the component will be served
        # by the local dev server that you run via `npm run start`.
        # (This is useful while your component is in development.)
        url="http://localhost:5173",
    )
else:
    # When we're distributing a production version of the component, we'll
    # replace the `url` param with `path`, and point it to to the component's
    # build directory:
    parent_dir = os.path.dirname(os.path.abspath(__file__))
    build_dir = os.path.join(parent_dir, "dist")
    _record_viewer = components.declare_component(
        "record_viewer", path=build_dir
    )


# Create a wrapper function for the component. This is an optional
# best practice - we could simply expose the component function returned by
# `declare_component` and call it done. The wrapper allows us to customize
# our component's API: we can pre-process its input args, post-process its
# output value, and add a docstring for users.
def record_viewer(record_json, app_json, key=None):
    """Create a new instance of "record_viewer", which produces a timeline

    Args:
        record_json (object): JSON of the record serialized by `json.loads`.
        app_json (object): JSON of the app serialized by `json.loads`.

    Returns:
        string: Start time of the selected component in the application. If the whole app is selected,
    """

    # Call through to our private component function. Arguments we pass here
    # will be sent to the frontend, where they'll be available in an "args"
    # dictionary.
    #
    # "default" is a special argument that specifies the initial return
    # value of the component before the user has interacted with it.
    component_value = _record_viewer(
        record_json=record_json, app_json=app_json, key=key, default=""
    )

    return component_value



import asyncio
from pprint import PrettyPrinter
from threading import Thread
from typing import Callable, List, Mapping, Optional, Sequence, Union

from ipywidgets import widgets
import traitlets
from traitlets import HasTraits
from traitlets import Unicode
from trulens.core import app as mod_app
from trulens.core.instruments import Instrument
from trulens.core.utils.json import JSON_BASES
from trulens.core.utils.json import jsonify_for_ui
from trulens.core.utils.serial import Lens

pp = PrettyPrinter()

debug_style = dict(border="0px solid gray", padding="0px")

VALUE_MAX_CHARS = 1024


class Selector(HasTraits):
    select = Unicode()
    jpath = traitlets.Any()

    def __init__(self, select: Union[Lens, str], make_on_delete: Callable):
        if isinstance(select, Lens):
            self.select = str(select)
            self.jpath = select
        else:
            self.select = select
            self.jpath = Lens.of_string(select)

        self.w_edit = widgets.Text(value=select, layout=debug_style)
        self.w_delete = widgets.Button(
            description="x", layout=dict(width="30px", **debug_style)
        )

        self.on_delete = make_on_delete(self)
        self.w_delete.on_click(self.on_delete)

        traitlets.link((self.w_edit, "value"), (self, "select"))

        def on_update_select(ev):
            try:
                jpath = Lens.of_string(ev.new)
                self.jpath = jpath
                self.w_edit.layout.border = "0px solid black"
            except Exception:
                self.w_edit.layout.border = "1px solid red"

        self.observe(on_update_select, ["select"])

        self.w = widgets.HBox([self.w_delete, self.w_edit], layout=debug_style)


class SelectorValue(HasTraits):
    selector = traitlets.Any()
    obj = traitlets.Any()

    def __init__(
        self,
        selector: Selector,
        stdout_display: widgets.Output,
        instrument: Instrument,
    ):
        self.selector = selector
        self.obj = None

        self.stdout_display = stdout_display

        self.w_listing = widgets.HTML(layout=debug_style)
        self.w = widgets.VBox(
            [self.selector.w, self.w_listing], layout=debug_style
        )

        self.selector.observe(self.update_selector, "jpath")
        self.observe(self.update_obj, "obj")

        self.instrument = instrument

    def _jsonify(self, obj):
        return jsonify_for_ui(obj=obj, instrument=self.instrument)

    def update_selector(self, ev):
        self.update()

    def update_obj(self, ev):
        self.update()

    def update(self):
        obj = self.obj
        jpath = self.selector.jpath

        inner_obj = None
        inner_class = None

        if obj is None:
            ret_html = "no listing yet"
        else:
            with self.stdout_display:
                try:
                    ret_html = ""

                    for inner_obj in jpath.get(obj):
                        inner_class = type(inner_obj)
                        inner_obj_id = id(inner_obj)
                        inner_obj = self._jsonify(inner_obj)

                        ret_html += f"<div>({inner_class.__name__} at 0x{inner_obj_id:x}): "  # as {type(inner_obj).__name__}): "

                        # if isinstance(inner_obj, pydantic.BaseModel):
                        #    inner_obj = inner_obj.model_dump()

                        if isinstance(inner_obj, JSON_BASES):
                            ret_html += str(inner_obj)[0:VALUE_MAX_CHARS]

                        elif isinstance(inner_obj, Mapping):
                            ret_html += "<ul>"
                            for key, val in inner_obj.items():
                                ret_html += f"<li>{key} = {str(val)[0:VALUE_MAX_CHARS]}</li>"
                            ret_html += "</ul>"

                        elif isinstance(inner_obj, Sequence):
                            ret_html += "<ul>"
                            for i, val in enumerate(inner_obj):
                                ret_html += f"<li>[{i}] = {str(val)[0:VALUE_MAX_CHARS]}</li>"
                            ret_html += "</ul>"

                        else:
                            ret_html += str(inner_obj)[0:VALUE_MAX_CHARS]

                        ret_html += "</div>"

                except Exception:
                    self.w_listing.layout.border = "1px solid red"
                    return

        self.w_listing.layout.border = "0px solid black"
        self.w_listing.value = f"<div>{ret_html}</div>"


class RecordWidget:
    def __init__(
        self,
        record_selections,
        instrument: Instrument,
        record=None,
        human_or_input=None,
        stdout_display: widgets.Output = None,
    ):
        self.record = record
        self.record_selections = record_selections
        self.record_values = dict()

        self.human_or_input = widgets.HBox([human_or_input], layout=debug_style)
        self.w_human = widgets.HBox(
            [widgets.HTML("<b>human:</b>"), self.human_or_input],
            layout=debug_style,
        )
        self.d_comp = widgets.HTML(layout=debug_style)
        self.d_extras = widgets.VBox(layout=debug_style)

        self.stdout_display = stdout_display

        self.human = ""
        self.comp = ""

        self.instrument = instrument

        self.d = widgets.VBox(
            [self.w_human, self.d_comp, self.d_extras],
            layout={**debug_style, "border": "5px solid #aaaaaa"},
        )

    def update_selections(self):
        # change to trait observe
        for s in self.record_selections:
            if s not in self.record_values:
                sv = SelectorValue(
                    selector=s,
                    stdout_display=self.stdout_display,
                    instrument=self.instrument,
                )
                self.record_values[s] = sv
                self.d_extras.children += (sv.w,)

            if self.record is not None:
                record_filled = self.record.layout_calls_as_app()
            else:
                record_filled = None

            self.record_values[s].obj = record_filled

    def remove_selector(self, selector: Selector):
        if selector not in self.record_values:
            return

        item = self.record_values[selector]
        del self.record_values[selector]
        new_children = list(self.d_extras.children)
        new_children.remove(item.w)
        self.d_extras.children = tuple(new_children)

    def set_human(self, human: str):
        self.human = human
        self.human_or_input.children = (
            widgets.HTML(f"<div>{human}</div>", layout=debug_style),
        )

    def set_comp(self, comp: str):
        self.comp = comp
        self.d_comp.value = f"<div><b>computer:</b> {comp}</div>"


class AppUI(traitlets.HasTraits):
    # very prototype

    def __init__(
        self,
        app: mod_app.App,
        use_async: bool = False,
        app_selectors: Optional[List[Union[str, Lens]]] = None,
        record_selectors: Optional[List[Union[str, Lens]]] = None,
    ):
        self.use_async = use_async

        self.app = app

        self.main_input = widgets.Text(layout=debug_style)
        self.app_selector = widgets.Text(layout=debug_style)
        self.record_selector = widgets.Text(layout=debug_style)

        self.main_input_button = widgets.Button(
            description="+ Record", layout=debug_style
        )
        self.app_selector_button = widgets.Button(
            description="+ Select.App", layout=debug_style
        )
        self.record_selector_button = widgets.Button(
            description="+ Select.Record", layout=debug_style
        )

        self.display_top = widgets.VBox([], layout=debug_style)
        self.display_side = widgets.VBox(
            [], layout={"width": "50%", **debug_style}
        )

        self.display_stdout = widgets.Output()

        self.display_records = []

        self.app_selections = {}
        self.record_selections = []

        self.current_record = RecordWidget(
            record_selections=self.record_selections,
            human_or_input=self.main_input,
            stdout_display=self.display_stdout,
            instrument=self.app.instrument,
        )
        self.current_record_record = None

        self.records = [self.current_record]

        self.main_input.on_submit(self.add_record)
        self.app_selector.on_submit(self.add_app_selection)
        self.record_selector.on_submit(self.add_record_selection)

        self.main_input_button.on_click(self.add_record)
        self.app_selector_button.on_click(self.add_app_selection)
        self.record_selector_button.on_click(self.add_record_selection)

        outputs_widget = widgets.Accordion(children=[self.display_stdout])
        outputs_widget.set_title(0, "stdpipes")

        self.display_bottom = widgets.VBox(
            [
                widgets.HBox(
                    [self.main_input_button, self.main_input],
                    layout=debug_style,
                ),
                widgets.HBox(
                    [self.app_selector_button, self.app_selector],
                    layout=debug_style,
                ),
                widgets.HBox(
                    [self.record_selector_button, self.record_selector],
                    layout=debug_style,
                ),
            ],
            layout=debug_style,
        )

        self.display_top.children += (self.current_record.d,)

        self.widget = widgets.VBox([
            widgets.HBox(
                [
                    widgets.VBox(
                        [self.display_top, self.display_bottom],
                        layout={**debug_style, "width": "50%"},
                    ),
                    self.display_side,
                ],
                layout=debug_style,
            ),
            outputs_widget,
        ])

        if app_selectors is not None:
            for selector in app_selectors:
                self._add_app_selector(selector)

        if record_selectors is not None:
            for selector in record_selectors:
                self._add_record_selector(selector)

    def make_on_delete_record_selector(self, selector):
        def on_delete(ev):
            self.record_selections.remove(selector)

            for r in self.records:
                r.remove_selector(selector)

        return on_delete

    def make_on_delete_app_selector(self, selector):
        def on_delete(ev):
            sw = self.app_selections[selector]
            del self.app_selections[selector]

            new_children = list(self.display_side.children)
            new_children.remove(sw.w)

            self.display_side.children = tuple(new_children)

        return on_delete

    def update_app_selections(self):
        for _, sw in self.app_selections.items():
            sw.update()

    def _add_app_selector(self, selector: Union[Lens, str]):
        with self.display_stdout:
            sel = Selector(
                select=selector, make_on_delete=self.make_on_delete_app_selector
            )

        sw = SelectorValue(
            selector=sel,
            stdout_display=self.display_stdout,
            instrument=self.app.instrument,
        )
        self.app_selections[sel] = sw
        sw.obj = self.app

        self.display_side.children += (sw.w,)

    def add_app_selection(self, w):
        self._add_app_selector(self.app_selector.value)

    def _add_record_selector(self, selector: Union[Lens, str]):
        with self.display_stdout:
            sel = Selector(
                select=selector,
                make_on_delete=self.make_on_delete_record_selector,
            )

        self.record_selections.append(sel)

        for r in self.records:
            r.update_selections()

    def add_record_selection(self, w):
        s = self.record_selector.value

        self._add_record_selector(s)

    def add_record(self, w):
        human = self.main_input.value

        if len(human) == 0:
            return

        self.current_record.set_human(human)

        with self.app as recording:
            # generalize
            if self.use_async:
                self.current_record.set_comp("generating:")

                comp = ""

                def run_in_thread(comp):
                    async def run_in_main_loop(comp):
                        comp_generator = await self.app.main_acall(human)
                        async for tok in comp_generator:
                            comp += tok
                            self.current_record.set_comp(comp)

                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    loop.run_until_complete(
                        asyncio.Task(run_in_main_loop(comp))
                    )

                t = Thread(target=run_in_thread, args=(comp,))
                t.start()
                t.join()

            else:
                with self.display_stdout:
                    self.current_record.set_comp("...")
                    comp = self.app.main_call(human)
                    self.current_record.set_comp(comp)

        self.current_record_record = recording.get()
        self.current_record.record = self.current_record_record
        self.current_record.update_selections()

        self.update_app_selections()

        self.current_record = RecordWidget(
            record_selections=self.record_selections,
            human_or_input=self.main_input,
            stdout_display=self.display_stdout,
            instrument=self.app.instrument,
        )
        self.records.append(self.current_record)
        self.display_top.children += (self.current_record.d,)



from importlib.metadata import version

from trulens.core.utils.imports import safe_importlib_package_name
from trulens.dashboard.run import run_dashboard
from trulens.dashboard.run import stop_dashboard

__version__ = version(safe_importlib_package_name(__package__ or __name__))


__all__ = [
    "run_dashboard",
    "stop_dashboard",
]



import asyncio
import json
import pprint as pp
import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from st_aggrid import AgGrid
from st_aggrid.grid_options_builder import GridOptionsBuilder
from st_aggrid.shared import GridUpdateMode
from st_aggrid.shared import JsCode
import streamlit as st
from streamlit_pills import pills
from trulens.core import TruSession
from trulens.core.app import LLM
from trulens.core.app import Agent
from trulens.core.app import ComponentView
from trulens.core.app import Other
from trulens.core.app import Prompt
from trulens.core.app import Tool
from trulens.core.database.base import MULTI_CALL_NAME_DELIMITER
from trulens.core.schema.select import Select
from trulens.core.utils.json import jsonify_for_ui
from trulens.core.utils.serial import Lens
from trulens.dashboard.components.record_viewer import record_viewer
from trulens.dashboard.streamlit_utils import init_from_args
from trulens.dashboard.ux.components import draw_agent_info
from trulens.dashboard.ux.components import draw_llm_info
from trulens.dashboard.ux.components import draw_prompt_info
from trulens.dashboard.ux.components import draw_tool_info
from trulens.dashboard.ux.components import render_selector_markdown
from trulens.dashboard.ux.page_config import set_page_config
from trulens.dashboard.ux.styles import CATEGORY
from trulens.dashboard.ux.styles import cellstyle_jscode

# https://github.com/jerryjliu/llama_index/issues/7244:
asyncio.set_event_loop(asyncio.new_event_loop())

set_page_config(page_title="Evaluations")
st.title("Evaluations")

if __name__ == "__main__":
    # If not imported, gets args from command line and creates Tru singleton
    init_from_args()

session = TruSession()
lms = session.connector.db

# TODO: remove code redundancy / redundant database calls
feedback_directions = {
    (
        row.feedback_json.get("supplied_name", "")
        or row.feedback_json["implementation"]["name"]
    ): (
        "HIGHER_IS_BETTER"
        if row.feedback_json.get("higher_is_better", True)
        else "LOWER_IS_BETTER"
    )
    for _, row in lms.get_feedback_defs().iterrows()
}
default_direction = "HIGHER_IS_BETTER"


def render_component(
    query: Lens, component: ComponentView, header: bool = True
) -> None:
    """Render the accessor/path within the wrapped app of the component."""

    if header:
        st.markdown(
            f"##### Component {render_selector_markdown(Select.for_app(query))}"
        )

    # Draw the python class information of this component.
    cls = component.cls
    base_cls = cls.base_class()
    label = f"__{repr(cls)}__"
    if str(base_cls) != str(cls):
        label += f" < __{repr(base_cls)}__"
    st.write("Python class: " + label)

    # Per-component-type drawing routines.
    if isinstance(component, LLM):
        draw_llm_info(component=component, query=query)

    elif isinstance(component, Prompt):
        draw_prompt_info(component=component, query=query)

    elif isinstance(component, Agent):
        draw_agent_info(component=component, query=query)

    elif isinstance(component, Tool):
        draw_tool_info(component=component, query=query)

    elif isinstance(component, Other):
        with st.expander("Uncategorized Component Details:"):
            st.json(jsonify_for_ui(component.json))

    else:
        with st.expander("Unhandled Component Details:"):
            st.json(jsonify_for_ui(component.json))


def render_record_metrics(
    app_df: pd.DataFrame, selected_rows: pd.DataFrame
) -> None:
    """Render record level metrics (e.g. total tokens, cost, latency) compared
    to the average when appropriate."""

    app_specific_df = app_df[app_df["app_id"] == selected_rows["app_id"][0]]

    token_col, cost_col, latency_col = st.columns(3)

    num_tokens = selected_rows["total_tokens"][0]
    token_col.metric(label="Total tokens (#)", value=num_tokens)

    cost = selected_rows["total_cost"][0]
    average_cost = app_specific_df["total_cost"].mean()
    delta_cost = f"{cost - average_cost:.3g}"
    cost_col.metric(
        label="Total cost (USD)",
        value=selected_rows["total_cost"][0],
        delta=delta_cost,
        delta_color="inverse",
    )

    latency = selected_rows["latency"][0]
    average_latency = app_specific_df["latency"].mean()
    delta_latency = f"{latency - average_latency:.3g}s"
    latency_col.metric(
        label="Latency (s)",
        value=selected_rows["latency"][0],
        delta=delta_latency,
        delta_color="inverse",
    )


def extract_metadata(row: pd.Series) -> str:
    """Extract metadata from the record_json and return the metadata as a string.

    Args:
        row: The row containing the record_json.

    Returns:
        str: The metadata extracted from the record_json.
    """
    record_data = json.loads(row["record_json"])
    return str(record_data["meta"])


app_ids = list(app["app_id"] for app in lms.get_apps())


def get_apps():
    return list(lms.get_apps())


apps = get_apps()

if "app" in st.session_state:
    app = st.session_state.app
else:
    app = app_ids

st.query_params["app"] = app

app_name_versions = list(
    set(f"{app['app_name']} - {app['app_version']}" for app in apps)
)

if "app" in st.session_state:
    selected_apps = st.session_state.app
    filtered_apps = [app for app in apps if app["app_id"] in selected_apps]
    app_name_versions = list(
        set(
            f"{app['app_name']} - {app['app_version']}" for app in filtered_apps
        )
    )
selected_app_name_versions = st.multiselect(
    "Filter Apps", app_name_versions, default=app_name_versions
)

selected_apps = [
    app["app_id"]
    for app in apps
    if f"{app['app_name']} - {app['app_version']}" in selected_app_name_versions
]

df_results, feedback_cols = lms.get_records_and_feedback(app_ids=selected_apps)

if len(selected_apps) == 0:
    st.header("All Applications")

elif len(selected_apps) == 1:
    st.header(selected_app_name_versions[0])

else:
    st.header("Multiple Applications Selected")

if df_results.empty:
    st.write("No records yet...")
else:
    app_df = df_results

    tab1, tab2 = st.tabs(["Records", "Feedback Functions"])

    with tab1:
        gridOptions = {"alwaysShowHorizontalScroll": True}
        evaluations_df = app_df

        # By default the cells in the df are unicode-escaped, so we have to reverse it.
        input_array = evaluations_df["input"].to_numpy()
        output_array = evaluations_df["output"].to_numpy()

        # pull additional data from app_json into columns
        evaluations_df["app_name"] = evaluations_df["app_json"].apply(
            lambda x: json.loads(x)["app_name"]
        )
        evaluations_df["app_version"] = evaluations_df["app_json"].apply(
            lambda x: json.loads(x)["app_version"]
        )

        metadata = json.loads(evaluations_df["app_json"][0])["metadata"]
        metadata_cols = list(metadata.keys())
        for key in metadata_cols:
            evaluations_df[key] = evaluations_df["app_json"].apply(
                lambda x: json.loads(x)["metadata"].get(key)
            )

        feedback_cost_cols = []

        for col in evaluations_df.columns:
            if "feedback cost" in col:
                feedback_cost_cols.append(col)

        # Drop the original app_metadata column if it exists
        if "app_metadata" in evaluations_df.columns:
            evaluations_df.drop(columns=["app_metadata"], inplace=True)

        # Reorder columns
        columns = (
            ["app_name", "app_version", "tags"]
            + metadata_cols
            + [
                col
                for col in evaluations_df.columns
                if col
                not in (["app_name", "app_version", "tags"] + metadata_cols)
            ]
        )
        evaluations_df = evaluations_df[columns]

        decoded_input = np.vectorize(
            lambda x: x.encode("utf-8").decode("unicode-escape")
        )(input_array)
        decoded_output = np.vectorize(
            lambda x: x.encode("utf-8").decode("unicode-escape")
        )(output_array)

        evaluations_df["input"] = decoded_input
        evaluations_df["output"] = decoded_output

        # Apply the function to each row and create a new column 'record_metadata'
        evaluations_df["record_metadata"] = evaluations_df.apply(
            extract_metadata, axis=1
        )

        gb = GridOptionsBuilder.from_dataframe(evaluations_df)

        gb.configure_column("app_name", header_name="App Name")
        gb.configure_column("app_version", header_name="App Version")
        gb.configure_column("type", header_name="App Type", hide=True)
        gb.configure_column("record_json", header_name="Record JSON", hide=True)
        gb.configure_column("app_json", header_name="App JSON", hide=True)
        gb.configure_column("cost_json", header_name="Cost JSON", hide=True)
        gb.configure_column("perf_json", header_name="Perf. JSON", hide=True)

        gb.configure_column("record_id", header_name="Record ID", hide=True)
        gb.configure_column("app_id", header_name="App ID", hide=True)

        gb.configure_column("feedback_id", header_name="Feedback ID", hide=True)
        gb.configure_column("input", header_name="User Input")
        gb.configure_column("output", header_name="Response")
        gb.configure_column("record_metadata", header_name="Record Metadata")

        gb.configure_column("total_tokens", header_name="Total Tokens (#)")
        gb.configure_column("total_cost", header_name="Total Cost (App)")
        gb.configure_column("cost_currency", header_name="Cost Currency")
        gb.configure_column("latency", header_name="Latency (Seconds)")
        gb.configure_column("tags", header_name="Application Tag")
        gb.configure_column("ts", header_name="Time Stamp", sort="desc")

        non_feedback_cols = (
            [
                "app_name",
                "app_version",
                "app_id",
                "type",
                "ts",
                "total_tokens",
                "total_cost",
                "cost_currency",
                "record_json",
                "latency",
                "tags",
                "record_metadata",
                "record_id",
                "cost_json",
                "app_json",
                "input",
                "output",
                "perf_json",
            ]
            + metadata_cols
            + feedback_cost_cols
        )

        for feedback_col in evaluations_df.columns.drop(non_feedback_cols):
            if "distance" in feedback_col:
                gb.configure_column(
                    feedback_col, hide=feedback_col.endswith("_calls")
                )
            else:
                # cell highlight depending on feedback direction
                cellstyle = JsCode(
                    cellstyle_jscode[
                        feedback_directions.get(feedback_col, default_direction)
                    ]
                )

                gb.configure_column(
                    feedback_col,
                    cellStyle=cellstyle,
                    hide=feedback_col.endswith("_calls"),
                )

        gb.configure_pagination()
        gb.configure_side_bar()
        gb.configure_selection(selection_mode="single", use_checkbox=False)

        gridOptions = gb.build()
        data = AgGrid(
            evaluations_df,
            gridOptions=gridOptions,
            update_mode=GridUpdateMode.SELECTION_CHANGED,
            allow_unsafe_jscode=True,
        )

        selected_rows = data.selected_rows
        selected_rows = pd.DataFrame(selected_rows)

        if len(selected_rows) == 0:
            st.write("Hint: select a row to display details of a record")

        else:
            # Start the record specific section
            st.divider()

            # Breadcrumbs
            st.caption(
                f"{selected_rows['app_id'][0]} / {selected_rows['record_id'][0]}"
            )
            st.header(f"{selected_rows['record_id'][0]}")

            render_record_metrics(app_df, selected_rows)

            st.markdown("")

            prompt = selected_rows["input"][0]
            response = selected_rows["output"][0]
            details = selected_rows["app_json"][0]
            record_str = selected_rows["record_json"][0]
            record_json = json.loads(record_str)
            record_metadata = selected_rows["record_metadata"][0]

            app_json = json.loads(
                details
            )  # apps may not be deserializable, don't try to, keep it json.

            row = selected_rows.head().iloc[0]

            st.markdown("#### Feedback results")
            if len(feedback_cols) == 0:
                st.write("No feedback details")
            else:
                feedback_with_valid_results = sorted(
                    list(
                        filter(
                            lambda fcol: row[fcol] is not None, feedback_cols
                        )
                    )
                )

                def get_icon(feedback_name):
                    cat = CATEGORY.of_score(
                        row[feedback_name],
                        higher_is_better=feedback_directions.get(
                            feedback_name, default_direction
                        )
                        == default_direction,
                    )
                    return cat.icon

                icons = list(
                    map(
                        lambda fcol: get_icon(fcol), feedback_with_valid_results
                    )
                )

                selected_fcol = None
                if len(feedback_with_valid_results) > 0:
                    selected_fcol = pills(
                        "Feedback functions (click on a pill to learn more)",
                        feedback_with_valid_results,
                        index=None,
                        format_func=lambda fcol: f"{fcol} {row[fcol]:.4f}",
                        icons=icons,
                    )
                else:
                    st.write("No feedback functions found.")

                def display_feedback_call(call, feedback_name):
                    def highlight(s):
                        if "distance" in feedback_name:
                            return [
                                f"background-color: {CATEGORY.UNKNOWN.color}"
                            ] * len(s)
                        cat = CATEGORY.of_score(
                            s.result,
                            higher_is_better=feedback_directions.get(
                                feedback_name, default_direction
                            )
                            == default_direction,
                        )
                        return [f"background-color: {cat.color}"] * len(s)

                    def highlight_groundedness(s):
                        if "distance" in feedback_name:
                            return [
                                f"background-color: {CATEGORY.UNKNOWN.color}"
                            ] * len(s)
                        cat = CATEGORY.of_score(
                            s.Score,
                            higher_is_better=feedback_directions.get(
                                feedback_name, default_direction
                            )
                            == default_direction,
                        )
                        return [f"background-color: {cat.color}"] * len(s)

                    if call is not None and len(call) > 0:
                        # NOTE(piotrm for garett): converting feedback
                        # function inputs to strings here as other
                        # structures get rendered as [object Object] in the
                        # javascript downstream. If the first input/column
                        # is a list, the DataFrame.from_records does create
                        # multiple rows, one for each element, but if the
                        # second or other column is a list, it will not do
                        # this.
                        for c in call:
                            args = c["args"]
                            for k, v in args.items():
                                if not isinstance(v, str):
                                    args[k] = pp.pformat(v)

                        df = pd.DataFrame.from_records(c["args"] for c in call)

                        df["result"] = pd.DataFrame([
                            float(call[i]["ret"])
                            if call[i]["ret"] is not None
                            else -1
                            for i in range(len(call))
                        ])
                        df["meta"] = pd.Series([
                            call[i]["meta"] for i in range(len(call))
                        ])
                        df = df.join(
                            df.meta.apply(lambda m: pd.Series(m))
                        ).drop(columns="meta")

                        # note: improve conditional to not rely on the feedback name
                        if "groundedness" in feedback_name.lower():
                            try:
                                # Split the reasons value into separate rows and columns
                                reasons = df["reasons"].iloc[0]
                                # Split the reasons into separate statements
                                statements = reasons.split("STATEMENT")
                                data = []
                                # Each reason has three components: statement, supporting evidence, and score
                                # Parse each reason into these components and add them to the data list
                                for statement in statements[1:]:
                                    try:
                                        criteria = statement.split(
                                            "Criteria: "
                                        )[1].split("Supporting Evidence: ")[0]
                                        supporting_evidence = statement.split(
                                            "Supporting Evidence: "
                                        )[1].split("Score: ")[0]
                                        score_pattern = re.compile(
                                            r"([0-9]+)(?=\D*$)"
                                        )
                                        score_split = statement.split(
                                            "Score: "
                                        )[1]
                                        score_match = score_pattern.search(
                                            score_split
                                        )
                                        if score_match:
                                            score = (
                                                float(score_match.group(1)) / 10
                                            )
                                    except Exception:
                                        pass
                                    data.append({
                                        "Statement": criteria,
                                        "Supporting Evidence from Source": supporting_evidence,
                                        "Score": score,
                                    })
                                reasons_df = pd.DataFrame(data)
                                # Combine the original feedback data with the expanded reasons
                                df_expanded = pd.concat(
                                    [
                                        df.reset_index(drop=True),
                                        reasons_df.reset_index(drop=True),
                                    ],
                                    axis=1,
                                )
                                st.dataframe(
                                    df_expanded.style.apply(
                                        highlight_groundedness, axis=1
                                    ).format("{:.2f}", subset=["Score"]),
                                    hide_index=True,
                                    column_order=[
                                        "Statement",
                                        "Supporting Evidence from Source",
                                        "Score",
                                    ],
                                )
                            except Exception:
                                st.dataframe(
                                    df.style.apply(highlight, axis=1),
                                    hide_index=True,
                                )
                        else:
                            st.dataframe(
                                df.style.apply(highlight, axis=1),
                                hide_index=True,
                            )
                    else:
                        st.text("No feedback details.")

                if selected_fcol is not None:
                    try:
                        if MULTI_CALL_NAME_DELIMITER in selected_fcol:
                            fcol = selected_fcol.split(
                                MULTI_CALL_NAME_DELIMITER
                            )[0]
                        feedback_calls = row[f"{selected_fcol}_calls"]
                        display_feedback_call(feedback_calls, selected_fcol)
                    except Exception:
                        pass

            st.subheader("Trace details")
            val = record_viewer(record_json, app_json)
            st.markdown("")

    with tab2:
        feedback = feedback_cols
        cols = 4
        rows = len(feedback) // cols + 1

        for row_num in range(rows):
            with st.container():
                columns = st.columns(cols)
                for col_num in range(cols):
                    with columns[col_num]:
                        ind = row_num * cols + col_num
                        if ind < len(feedback):
                            # Generate histogram
                            fig, ax = plt.subplots()
                            bins = [
                                0,
                                0.2,
                                0.4,
                                0.6,
                                0.8,
                                1.0,
                            ]  # Quintile buckets
                            ax.hist(
                                app_df[feedback[ind]],
                                bins=bins,
                                edgecolor="black",
                                color="#2D736D",
                            )
                            ax.set_xlabel("Feedback Value")
                            ax.set_ylabel("Frequency")
                            ax.set_title(feedback[ind], loc="center")
                            st.pyplot(fig)



import argparse
import sys

from trulens.core import TruSession
from trulens.core.database import base as mod_db


def init_from_args():
    """Parse command line arguments and initialize Tru with them.

    As Tru is a singleton, further TruSession() uses will get the same configuration.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--database-url", default=None)
    parser.add_argument(
        "--database-prefix", default=mod_db.DEFAULT_DATABASE_PREFIX
    )

    try:
        args = parser.parse_args()
    except SystemExit as e:
        print(e)

        # This exception will be raised if --help or invalid command line arguments
        # are used. Currently, streamlit prevents the program from exiting normally,
        # so we have to do a hard exit.
        sys.exit(e.code)

    TruSession(
        database_url=args.database_url, database_prefix=args.database_prefix
    )



import re
import time
from typing import Dict, List

import pandas as pd
from trulens.core.schema.feedback import FeedbackDefinition
from trulens.core.schema.record import Record
from trulens.dashboard.ux.styles import CATEGORY


def get_icon(fdef: FeedbackDefinition, result: float) -> str:
    """
    Get the icon for a given feedback definition and result.

    Args:
        fdef (FeedbackDefinition): The feedback definition
        result (float): The result of the feedback

    Returns:
        str: The icon for the feedback
    """
    cat = CATEGORY.of_score(
        result or 0,
        higher_is_better=fdef.higher_is_better
        if fdef.higher_is_better is not None
        else True,
    )
    return cat.icon


def get_feedback_result(
    tru_record: Record, feedback_name: str, timeout: int = 60
) -> pd.DataFrame:
    """
    Retrieve the feedback results for a given feedback name from a TruLens record.

    Args:
        tru_record: The record containing feedback and future results.
        feedback_name: The name of the feedback to retrieve results for.

    Returns:
        pd.DataFrame: A DataFrame containing the feedback results. If no feedback
            results are found, an empty DataFrame is returned.
    """
    start_time = time.time()
    feedback_calls = None

    while time.time() - start_time < timeout:
        feedback_calls = next(
            (
                future_result.result()
                for feedback_definition, future_result in tru_record.feedback_and_future_results
                if feedback_definition.name == feedback_name
            ),
            None,
        )
        if feedback_calls is not None:
            break
        time.sleep(1)  # Wait for 1 second before checking again

    if feedback_calls is None:
        raise TimeoutError(
            f"Feedback for '{feedback_name}' not available within {timeout} seconds."
        )

    # Ensure feedback_calls is iterable
    if not hasattr(feedback_calls, "__iter__"):
        raise ValueError("feedback_calls is not iterable")

    feedback_result = [
        {**call.model_dump()["args"], "ret": call.model_dump()["ret"]}
        for call in feedback_calls.calls
    ]
    return pd.DataFrame(feedback_result)


def highlight(
    row: pd.Series,
    selected_feedback: str,
    feedback_directions: Dict[str, bool],
    default_direction: str,
) -> List[str]:
    """
    Apply background color to the rows of a DataFrame based on the selected feedback.

    Args:
        row (pandas.Series): A row of the DataFrame to be highlighted.
        selected_feedback (str): The selected feedback to determine the background color.
        feedback_directions (dict): A dictionary mapping feedback names to their directions.
        default_direction (str): The default direction for feedback.

    Returns:
        list: A list of CSS styles representing the background color for each cell in the row.
    """
    if "distance" in selected_feedback:
        return [f"background-color: {CATEGORY.DISTANCE.color}"] * len(row)

    cat = CATEGORY.of_score(
        row["score"],
        higher_is_better=feedback_directions.get(
            selected_feedback, default_direction == "HIGHER_IS_BETTER"
        ),
    )
    # Apply the background color to the entire row
    return [f"background-color: {cat.color}"] * len(row)


def expand_groundedness_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Expand the groundedness DataFrame by splitting the reasons column into separate rows and columns.

    Args:
        df (pd.DataFrame): The groundedness DataFrame.

    Returns:
        pd.DataFrame: The expanded DataFrame.
    """
    # Split the reasons value into separate rows and columns
    reasons = df["reasons"].iloc[0]
    # Split the reasons into separate statements
    statements = reasons.split("STATEMENT")
    data = []
    # Each reason has three components: statement, supporting evidence, and score
    # Parse each reason into these components and add them to the data list
    for statement in statements[1:]:
        try:
            criteria = statement.split("Criteria: ")[1].split(
                "Supporting Evidence: "
            )[0]
            supporting_evidence = statement.split("Supporting Evidence: ")[
                1
            ].split("Score: ")[0]
            score_pattern = re.compile(r"([0-9]+)(?=\D*$)")
            score_split = statement.split("Score: ")[1]
            score_match = score_pattern.search(score_split)
            if score_match:
                score = float(score_match.group(1)) / 10
            else:
                score = None
        except IndexError:
            # Handle cases where the expected substrings are not found
            criteria = None
            supporting_evidence = None
            score = None
        data.append({
            "Statement": criteria,
            "Supporting Evidence from Source": supporting_evidence,
            "Groundedness Score": score,
        })
    reasons_df = pd.DataFrame(data)

    reasons_df.rename(columns={"Groundedness Score": "score"}, inplace=True)

    # Return only the expanded reasons DataFrame
    return reasons_df.reset_index(drop=True)
